BUILD_BIT='64bit'
BUILD_CPU='x86_64'
BUILD_DATE='2024-01-26'
BUILD_MACHINE='aarch64'
BUILD_OS='linux'
BUILD_TIME='02:02'
BUILT_BY='antoine'
BUILT_ON='localhost.localdomain'
COMPILER_VERSION='aarch64-unknown-linux-gnu-gcc (GCC) 12.1.0'
CYTHON_VERSION='3.0.8'
LINKER_VERSION='GNU ld (GNU Binutils) 2.38'
PYTHON_VERSION='3.11.6'
lib_gobject_introspection='1.78.1'
lib_gtk='3.24.41'
lib_gtk_x11='3.24.41'
lib_nvenc='10'
lib_nvfbc='6.0'
lib_py3cairo='1.25.1'
lib_pygobject='3.46.0'
lib_python3='3.11'
lib_vpx='1.14.0'
lib_x11='1.8.7'
lib_x264='0.164.3108M'
lib_x265='3.5'
lib_xcomposite='0.4.6'
lib_xdamage='1.1.6'
lib_xext='1.3.5'
lib_xfixes='6.0.1'
lib_xkbfile='1.1.2'
lib_xrandr='1.5.4'
lib_xtst='1.2.4'
